import fnmatch

from dataclasses import dataclass, fields
import os
from transformers.file_utils import WEIGHTS_NAME, CONFIG_NAME, url_to_filename, hf_bucket_url
from transformers.tokenization_utils import ADDED_TOKENS_FILE, SPECIAL_TOKENS_MAP_FILE, TOKENIZER_CONFIG_FILE, \
    PreTrainedTokenizer
from ..transformersx_base import log,join_path

try:
    import turbo_transformers

    __turbo_available = True
except:
    __turbo_available = False


def is_turbo_available():
    return __turbo_available


@dataclass(frozen=True)
class ModelType:
    bert: str = "bert"
    albert: str = "albert"
    electra: str = "electra"
    roberta: str = "roberta"
    xlnet: str = "xlnet"

    @staticmethod
    def names():
        return [f.name for f in fields(ModelType)]


@dataclass
class ModelMode:
    classification: str = 'classification'
    regression: str = 'regression'

    @staticmethod
    def names():
        return [f.name for f in fields(ModelMode)]


@dataclass
class ModelTaskType:
    base: str = 'base'
    pretrain: str = 'pretrain'
    lm_head: str = 'lm_head'
    qa: str = 'qa'
    qa_s: str = 'qa_s'
    seq_cls: str = 'seq_cls'
    seq_cls_x: str = 'seq_cls_x'
    token_cls: str = 'token_cls'
    multi_choice: str = 'multi_choice'
    next_seq: str = 'next_seq'

    @staticmethod
    def names():
        return [f.name for f in fields(ModelTaskType)]


def _check_and_rename_pretrained_model_file(pretrained_model_dir, model_id, file_name, use_cdn):
    target_file_path = join_path(pretrained_model_dir, model_id, file_name)
    if os.path.exists(target_file_path):
        return True
    file_url = hf_bucket_url(model_id, file_name, use_cdn=use_cdn)
    file_dir_path = join_path(pretrained_model_dir, model_id)
    url_file_name = url_to_filename(file_url)

    matching_files = [
        file
        for file in fnmatch.filter(os.listdir(file_dir_path), url_file_name + ".*")
        if not file.endswith(".json") and not file.endswith(".lock")
    ]
    if len(matching_files) > 0:
        found_file_name = join_path(file_dir_path, matching_files[-1])
        os.rename(found_file_name, target_file_path)
        return True
    return False


def _check_and_rename_pretrained_model(pretrained_model_dir, model_id, tokenizer: PreTrainedTokenizer):
    model_files = [CONFIG_NAME, WEIGHTS_NAME, ADDED_TOKENS_FILE, SPECIAL_TOKENS_MAP_FILE, TOKENIZER_CONFIG_FILE]
    model_files.extend(tokenizer.vocab_files_names.values())
    for file_name in model_files:
        if _check_and_rename_pretrained_model_file(pretrained_model_dir, model_id, file_name, False):
            continue
        _check_and_rename_pretrained_model_file(pretrained_model_dir, model_id, file_name, True)


@dataclass(init=True)
class TaskModel:
    model_type: str = None
    config: type = None
    model_path: str = None
    model_class: type = None
    tokenizer: type = None
    main_parameter: str = None

    def load(self, **kwargs):
        cache_dir = kwargs['cache_dir']
        _check_and_rename_pretrained_model(cache_dir, self.model_path, self.tokenizer)
        model_path = join_path(cache_dir, self.model_path)

        self.model_path = model_path
        config = self.config.from_pretrained(
            model_path,
            num_labels=kwargs['num_labels']
        )
        tokenizer = self.tokenizer.from_pretrained(model_path)
        unit_test = kwargs['unit_test']
        if not unit_test:
            model = self.model_class.from_pretrained(
                model_path,
                config=config
            )
        else:
            model = None
        return config, tokenizer, model

    def renew(self, model_class):
        return TaskModel(self.model_type, self.config, self.model_path,
                         model_class, self.tokenizer,
                         self.main_parameter)


def model_func(model_type, config, tokenizer, main_parameter=None):
    def generate_model(model_path):
        return TaskModel(model_type, config, model_path, None, tokenizer, main_parameter)

    return generate_model


class TaskModels:
    MODELS = {}
    MODEL_CLASSES = {}

    def all_base_models(self, language=None):
        return dict([(l, [model.renew(self.MODEL_CLASSES[ModelTaskType.base]) for model in models]) for l, models in
                     self.MODELS.items() if language is None or l == language])

    def all_paths(self, language=None):
        return dict([(l, [model.model_path for model in models]) for l, models in self.MODELS.items() if
                     language is None or language == l])

    def _model_class(self, model_path, model_task_type, language='cn'):
        models_l = self.MODELS[language]
        if not models_l:
            return None
        for model in models_l:
            if model.model_path == model_path:
                model_class = self.MODEL_CLASSES[model_task_type]
                if not model_class:
                    return None
                return model.renew(model_class)

        return None

    def task_model(self, model_path, model_task_type, language='cn', ignore_not_exists=False) -> TaskModel:
        model = self._model_class(model_path, model_task_type, language)
        if not model and not ignore_not_exists:
            raise ValueError(
                "Cannot find the model path {} for language {},model_type={}".format(model_path, language,
                                                                                     model_task_type))
        return model



import os
import time

import torch
from torch.utils.data import Dataset
from tqdm.auto import tqdm, trange
from transformers import PreTrainedTokenizer, torch_distributed_zero_first, \
    RobertaTokenizer, RobertaTokenizerFast, XLMRobertaTokenizer
from typing import List, Optional, Union

from transformers.tokenization_utils import BatchEncoding

from .data_args import DataArguments
from .dataprocessor import DataProcessor

###TODO: The Dataset should be refactored for many data sources
###TODO: Behind the dataset, it should have a data sources. So the data processor should be the data sources.
from ..train.trainer_utils import InputFeatures, InputExample
from ai_harness import harnessutils as aiutils

log = aiutils.getLogger('task')


class TaskDataset(Dataset):
    """
    This will be superseded by a framework-agnostic approach
    soon.
    """

    args: DataArguments
    features: List[InputFeatures]

    def __init__(
            self,
            args: DataArguments,
            tokenizer: PreTrainedTokenizer,
            processor: DataProcessor,
            limit_length: Optional[int] = None,
            evaluate=False,
            local_rank=-1
    ):
        self.args = args
        # Load data features from cache or dataset file
        cached_features_file = None
        if not args.predict:
            cached_features_file = os.path.join(
                processor.data_dir(),
                "cached_{}_{}_{}_{}".format(
                    "dev" if evaluate else "train", tokenizer.__class__.__name__, str(args.max_seq_length),
                    args.task_name,
                ),
            )
        with torch_distributed_zero_first(local_rank):
            # Make sure only the first process in distributed training processes the dataset,
            # and the others will use the cache.

            if not args.predict and os.path.exists(cached_features_file) and not args.overwrite_cache:
                start = time.time()
                self.features = torch.load(cached_features_file)
                log.info(
                    f"Loading features from cached file {cached_features_file} [took %.3f s]", time.time() - start
                )
            else:
                log.info(f"Creating features from dataset file at {processor.data_dir()}")
                label_list = processor.get_labels()
                if args.task_name in ["mnli", "mnli-mm"] and tokenizer.__class__ in (
                        RobertaTokenizer,
                        RobertaTokenizerFast,
                        XLMRobertaTokenizer,
                ):
                    # HACK(label indices are swapped in RoBERTa pretrained model)
                    label_list[1], label_list[2] = label_list[2], label_list[1]
                examples = (
                    processor.get_dev_examples()
                    if evaluate
                    else processor.get_train_examples()
                )
                if limit_length is not None:
                    examples = examples[:limit_length]
                self.features = _glue_convert_examples_to_features(
                    examples,
                    tokenizer,
                    processor,
                    max_length=args.max_seq_length,
                    label_list=label_list,
                    output_mode=self.args.model_mode_for_data,
                    progress_bar=args.progress_bar,
                    evaluate=evaluate
                )
                if not args.predict and local_rank in [-1, 0]:
                    log.info("Saving features into cached file %s", cached_features_file)
                    start = time.time()
                    torch.save(self.features, cached_features_file)
                    # ^ This seems to take a lot of time so I want to investigate why and how we can improve.
                    log.info(
                        f"Saved features into cached file %s [took %.3f s]", cached_features_file, time.time() - start
                    )

    def __len__(self):
        return len(self.features)

    def __getitem__(self, i) -> InputFeatures:
        return self.features[i]


def _glue_convert_examples_to_features(
        examples: List[InputExample],
        tokenizer: PreTrainedTokenizer,
        processor: DataProcessor,
        max_length: Optional[int] = None,
        task=None,
        label_list=None,
        output_mode=None,
        progress_bar=False,
        evaluate=False
):
    """
        Loads a data file into a list of ``InputFeatures``

        Args:
            examples: List of ``InputExamples`` or ``tf.data.Dataset`` containing the examples.
            tokenizer: Instance of a tokenizer that will tokenize the examples
            max_length: Maximum example length. Defaults to the tokenizer's max_len
            task: GLUE task
            label_list: List of labels. Can be obtained from the processor using the ``processor.get_labels()`` method
            output_mode: String indicating the output mode. Either ``regression`` or ``classification``

        Returns:
            If the ``examples`` input is a ``tf.data.Dataset``, will return a ``tf.data.Dataset``
            containing the task-specific features. If the input is a list of ``InputExamples``, will return
            a list of task-specific ``InputFeatures`` which can be fed to the model.

        """
    log.info("Converting Examples to features .... total: " + str(len(examples)))
    if max_length is None:
        max_length = tokenizer.max_len

    has_label = (examples[0].label is not None)

    if has_label:
        if task is not None:
            if label_list is None:
                label_list = processor.get_labels()
                log.info("Using label list %s for task %s" % (label_list, task))

        label_map = {label: i for i, label in enumerate(label_list)}

        def label_from_example(example: InputExample) -> Union[int, float]:
            if output_mode == "classification":
                return label_map[example.label]
            elif output_mode == "regression":
                return float(example.label)
            raise KeyError(output_mode)

        labels = [label_from_example(example) for example in examples]

    # log.info("1. Tokenizer encoding examples .... total: " + str(len(examples)))
    # epoch_iterator = tqdm(examples, desc="Iteration", disable=not progress_bar)
    # batch_encoding = tokenizer.batch_encode_plus(
    #     [(example.text_a, example.text_b) for example in epoch_iterator], max_length=max_length, pad_to_max_length=True,
    # )

    batch_encoding = batch_encode_plus(tokenizer, examples, max_length, progress_bar)

    log.info("2. Converting Examples to Features .... total: " + str(len(examples)))
    epoch_iterator = tqdm(examples, desc="Iteration", disable=not progress_bar)
    features = []

    for k in batch_encoding.keys():
        log.info("key={},size={}".format(k, str(len(batch_encoding[k]))))

    for i, example in enumerate(epoch_iterator):
        inputs = {k: batch_encoding[k][i] for k in batch_encoding}
        label = labels[i] if has_label else None

        feature = InputFeatures(**inputs, label=label, guid=i if evaluate else None)

        features.append(feature)

    for i, example in enumerate(examples[:5]):
        log.info("*** Example ***")
        log.info("guid: %s" % (example.guid))
        log.info("features: %s" % features[i])

    return features


def batch_encode_plus(tokenizer, examples, max_length, progress_bar=False):
    log.info("1. Tokenizer encoding examples .... total: " + str(len(examples)))
    total = len(examples)
    epoch_iterator = tqdm(range(0, total, 100), desc="Iteration", disable=not progress_bar)

    batch_outputs = {}
    for step in epoch_iterator:
        batch_encoding = tokenizer.batch_encode_plus(
            [(example.text_a, example.text_b) for example in examples[step:step + 100]], max_length=max_length,
            pad_to_max_length=True,
        )

        for key, value in batch_encoding.items():
            if key not in batch_outputs:
                batch_outputs[key] = []
            batch_outputs[key].extend(value)

    return BatchEncoding(batch_outputs)



import json
from typing import Optional, List, Union

import dataclasses
from dataclasses import dataclass


@dataclass
class TaskInputExample:
    guid: str
    text_a: str
    text_b: Optional[str] = None
    label: Optional[str] = None

    def to_json_string(self):
        """Serializes this instance to a JSON string."""
        return json.dumps(dataclasses.asdict(self), indent=2) + "\n"


@dataclass(frozen=True)
class TaskInputFeatures:
    input_ids: List[int]
    attention_mask: Optional[List[int]] = None
    token_type_ids: Optional[List[int]] = None
    label: Optional[Union[int, float]] = None
    guid: int = None

    def to_json_string(self):
        """Serializes this instance to a JSON string."""
        return json.dumps(dataclasses.asdict(self)) + "\n"



import pandas as pd

from ..task_base import *


class SentimentDataProcessor(DataProcessor):
    def __init__(self, config: DataArguments):
        self._config = config

    def _get_example(self, file_name, type):
        pd_all = pd.read_csv(join_path(self._config.data_dir, file_name))

        log.info("Read data from {}, length={}".format(join_path(self._config.data_dir, file_name), len(pd_all)))
        examples = []
        for i, d in enumerate(pd_all.values):
            examples.append(InputExample(guid=type + '_' + str(i),
                                         text_a=d[1],
                                         label=str(d[0])))

        return examples

    def get_train_examples(self):
        return self._get_example('train.csv', 'train')

    def get_dev_examples(self):
        return self._get_example('dev.csv', 'dev')

    def get_labels(self):
        return ['0', '1', '2', '3']

    def data_dir(self):
        return self._config.data_dir


class SentimentTask(ExampleTaskBase):
    def __init__(self, taskArgs: TaskArguments = None):
        super().__init__('sentiment', taskArgs)
        self.task_args.model_args.num_labels = 4

    def _data_processor(self):
        return SentimentDataProcessor(self.task_args.data_args)



from ..task_base import *

import pandas as pd


class NewsTransformerDataProcessor(DataProcessor):
    def __init__(self, config: DataArguments):
        self._config = config

    def _get_example(self, file_name, type):
        pd_all = pd.read_csv(join_path(self._config.data_dir, file_name))

        log.info("Read data from {}, length={}".format(join_path(self._config.data_dir, file_name), len(pd_all)))
        return self._create_examples(pd_all, type)

    def _create_examples(self, lines, set_type):
        """Creates examples for the training and dev sets."""
        examples = []
        for (i, line) in enumerate(lines):
            guid = "%s-%s" % (set_type, i)
            text_a = line[1]
            label = line[0]
            examples.append(
                InputExample(guid=guid, text_a=text_a, text_b=None, label=label))
        return examples

    def get_train_examples(self):
        return self._get_example('train.txt', 'train')

    def get_dev_examples(self):
        return self._get_example('dev.txt', 'dev')

    def get_labels(self):
        """See base class."""
        return ["体育",
                "娱乐",
                "家居",
                "房产",
                "教育",
                "时尚",
                "时政",
                "游戏",
                "科技",
                "财经"]

    def data_dir(self):
        return self._config.data_dir


class SentimentTask(ExampleTaskBase):
    def __init__(self, taskArgs: TaskArguments = None):
        super().__init__('news_cls', taskArgs)
        super().task_args.model_args.num_labels = 10

    def _data_processor(self):
        return NewsTransformerDataProcessor(self.task_args.data_args)



from collections import OrderedDict

from transformers import is_tf_available
from transformers.modeling_xlnet import XLNetConfig

from transformers.tokenization_xlnet import XLNetTokenizer

from ..model_base import TaskModels, ModelType, ModelTaskType, model_func

default_model = model_func(ModelType.electra, XLNetConfig, XLNetTokenizer, 'xlnet')


class XLNet_Task_Models_Base(TaskModels):
    MODELS = {
        "cn": [default_model("clue/roberta_chinese_clue_tiny"),
               default_model("clue/roberta_chinese_3L312_clue_tiny"),
               default_model("roberta_chinese_3L768_clue_tiny"),
               default_model("clue/roberta_chinese_pair_tiny"),
               default_model("lonePatient/roberta_chinese_clue_tiny"),
               default_model("clue/roberta_chinese_clue_base"),
               default_model("hfl/chinese-roberta-wwm-ext"),
               default_model("clue/roberta_chinese_large"),
               default_model("chinese-roberta-wwm-ext-large"),
               default_model("clue/roberta_chinese_clue_large"),
               default_model("clue/roberta_chinese_pair_large")
               ]}


class XLNet_Task_Models(XLNet_Task_Models_Base):
    from transformers.modeling_xlnet import XLNetModel, XLNetForMultipleChoice, XLNetForQuestionAnswering, \
        XLNetForQuestionAnsweringSimple, XLNetForSequenceClassification, XLNetForTokenClassification, XLNetLMHeadModel
    MODEL_CLASSES = OrderedDict([
        (ModelTaskType.base, XLNetModel),
        (ModelTaskType.lm_head, XLNetLMHeadModel),
        (ModelTaskType.seq_cls, XLNetForSequenceClassification),
        (ModelTaskType.token_cls, XLNetForTokenClassification),
        (ModelTaskType.qa, XLNetForQuestionAnswering),
        (ModelTaskType.qa_s, XLNetForQuestionAnsweringSimple),
        (ModelTaskType.multi_choice, XLNetForMultipleChoice)
    ])


if is_tf_available():
    class TFXLNet_Task_Models(XLNet_Task_Models_Base):
        from transformers.modeling_tf_xlnet import TFXLNetModel, TFXLNetForQuestionAnsweringSimple, \
            TFXLNetForSequenceClassification, TFXLNetForTokenClassification, TFXLNetLMHeadModel
        MODEL_CLASSES = OrderedDict([
            (ModelTaskType.base, TFXLNetModel),
            (ModelTaskType.lm_head, TFXLNetLMHeadModel),
            (ModelTaskType.seq_cls, TFXLNetForSequenceClassification),
            (ModelTaskType.token_cls, TFXLNetForTokenClassification),
            (ModelTaskType.qa_s, TFXLNetForQuestionAnsweringSimple),
        ])



import torch
from torch import nn
from torch.nn import MSELoss, CrossEntropyLoss
from transformers import BertModel
import turbo_transformers
from turbo_transformers import PoolingType
from turbo_transformers import ReturnType


class TurboBertForSequenceClassification(nn.Module):
    def __init__(self, bert, config):
        super().__init__()
        self.num_labels = config.num_labels

        self.bert = bert
        self.dropout = nn.Dropout(config.hidden_dropout_prob)
        self.classifier = nn.Linear(config.hidden_size, config.num_labels)

    def forward(
            self,
            input_ids=None,
            attention_mask=None,
            token_type_ids=None,
            position_ids=None,
            head_mask=None,
            inputs_embeds=None,
            labels=None,
    ):

        outputs = self.bert(
            input_ids,
            attention_masks=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            pooling_type=PoolingType.FIRST,
            return_type=ReturnType.TORCH
        )

        pooled_output = outputs[1]

        pooled_output = self.dropout(pooled_output)
        logits = self.classifier(pooled_output)

        outputs = (logits,) + outputs[2:]  # add hidden states and attention if they are here

        if labels is not None:
            if self.num_labels == 1:
                #  We are doing regression
                loss_fct = MSELoss()
                loss = loss_fct(logits.view(-1), labels.view(-1))
            else:
                loss_fct = CrossEntropyLoss()
                loss = loss_fct(logits.view(-1, self.num_labels), labels.view(-1))
            outputs = (loss,) + outputs

        return outputs  # (loss), logits, (hidden_states), (attentions)

    @staticmethod
    def from_pretrained(model_id_or_path: str, config, **kwargs):
        torch_model = BertModel.from_pretrained(model_id_or_path, config=config, **kwargs)

        turbo_model = turbo_transformers.BertModelWithPooler.from_torch(torch_model, device=torch.device('cuda:0'))
        model = TurboBertForSequenceClassification(turbo_model, config)
        return model



from transformers import BertConfig, BertForSequenceClassification

from ai_transformersx import *


class TestTransformersModel:
    def test_task_model(self):
        model = task_model("bert-base-chinese", ModelTaskType.seq_cls, )
        assert issubclass(model.config, (BertConfig))
        assert issubclass(model.model_class, BertForSequenceClassification)

    def test_models(self):
        models = query_base_models(["bert", "albert"])
        assert models["bert"] and models["albert"]
        assert models["bert"]["cn"][0].model_path.index("bert") > -1



from ..model.model_args import ModelArguments
from ..data.data_args import DataArguments
from ..train.training_args import TrainingArguments
from ..transformersx_base import *


@configclass
class TaskArguments:
    action: str = field("train", "the task action: train,eval,predict")
    model_args: ModelArguments = ModelArguments()
    data_args: DataArguments = DataArguments()
    training_args: TrainingArguments = TrainingArguments()

    def validate_and_set_task_name(self, task_name):
        self.training_args.validate()
        if not task_name:
            self.training_args.output_dir = join_path(
                self.training_args.output_dir,
                self.model_args.model_name
            )
            return
        self.model_args.model_finetuned_dir = join_path(self.model_args.model_finetuned_dir, task_name)
        self.training_args.output_dir = join_path(self.training_args.output_dir, task_name,
                                                            self.model_args.model_name)


def parse_tasks_args(argsObjOrClass=None):
    if argsObjOrClass is None:
        task_args = TaskArguments()
    elif argsObjOrClass is not None and type(argsObjOrClass) != type:
        task_args = argsObjOrClass
    else:
        task_args = argsObjOrClass()

    return Arguments(task_args).parse()



from ai_harness.configclasses import configclass, field
from ai_harness.configuration import Arguments
from ai_harness.executors import QueueExecutor
from ai_harness.fileutils import join_path

from ai_transformersx import query_base_models, task_model


@configclass()
class DownloadConfiguration:
    model_type: str = field('bert', 'specified the model type')
    model_path: str = field('bert', 'specified the model path')
    language: str = field('cn', 'specified the language of model')
    cache_dir: str = field('nlp_models_cache', 'specified the cache dir for models')


class Downloader():
    def __init__(self, config: DownloadConfiguration):
        self._config = config
        self.all_models = []

    def __call__(self, *args, **kwargs):
        self.__download_single()

    def __download_single(self):
        self.__download_single_language("pt")
        self.__download_single_language("tf")

    def __download_single_language(self, language):
        model = task_model(model_path=self._config.model_path,
                           model_type=self._config.model_type,
                           model_task_type='base',
                           framework=self._config.framework,
                           language=language
                           )
        self._download_model(model)

    def __download_all(self, types):
        self.__build_models(types, 'pt')
        self.__build_models(types, 'tf')
        QueueExecutor(self.all_models, worker_num=4).run(self._download_models)

    def __build_models(self, types, framework):
        models = query_base_models(types=types.split(','),
                                   framework=framework,
                                   language=self._config.language)
        if not models:
            return
        for models in models.values():
            self.all_models.extend(models[self._config.language])

    def _download_models(self, models):
        [self._download_model(model) for model in models]

    def _download_model(self, model):
        cache_dir = join_path(self._config.cache_dir, model.model_path)
        self.__from_pretrained(model.config, model.model_path, cache_dir)
        self.__from_pretrained(model.tokenizer, model.model_path, cache_dir)
        self.__from_pretrained(model.model_class, model.model_path, cache_dir)

    def __from_pretrained(self, cls, path, cache_dir):
        print("Loading from {} with {}".format(path, str(cls)))
        try:
            cls.from_pretrained(path, cache_dir=cache_dir)
        except Exception as ex:
            print("Failed to load from {} with {}, error: ".format(path, str(cls), str(ex)))
            pass


if __name__ == "__main__":
    config: DownloadConfiguration = Arguments(DownloadConfiguration()).parse()

    Downloader(config)()
